# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import threading

try:
    from unittest import mock
except ImportError:  # pragma: NO PY3 COVER
    import mock  # type: ignore

import pytest  # type: ignore

from test_utils import orchestrate


def test__permutations():
    sequence = [1, 2, 3, 1, 2, 3, 1, 2, 3]
    permutations = orchestrate._permutations(sequence)
    assert len(permutations) == 1680

    result = list(permutations)
    assert len(permutations) == len(result)  # computed length matches reality
    assert len(result) == len(set(result))  # no duplicates
    assert result[0] == (1, 1, 1, 2, 2, 2, 3, 3, 3)
    assert result[-1] == (3, 3, 3, 2, 2, 2, 1, 1, 1)

    assert list(orchestrate._permutations([1, 2, 3])) == [
        (1, 2, 3),
        (1, 3, 2),
        (2, 1, 3),
        (2, 3, 1),
        (3, 1, 2),
        (3, 2, 1),
    ]


class Test_orchestrate:
    @staticmethod
    def test_bad_keyword_argument():
        with pytest.raises(TypeError):
            orchestrate.orchestrate(None, None, what="for?")

    @staticmethod
    def test_no_failures():
        test_calls = []

        def make_test(name):
            def test():  # pragma: NO COVER
                test_calls.append(name)  # pragma: SYNCPOINT
                test_calls.append(name)  # pragma: SYNCPOINT
                test_calls.append(name)

            return test

        test1 = make_test("A")
        test2 = make_test("B")

        permutations = orchestrate._permutations(["A", "B", "A", "B", "A", "B"])
        expected = list(itertools.chain(*permutations))

        counts = orchestrate.orchestrate(test1, test2)
        assert counts == (3, 3)
        assert test_calls == expected

    @staticmethod
    def test_named_syncpoints():
        test_calls = []

        def make_test(name):
            def test():  # pragma: NO COVER
                test_calls.append(name)  # pragma: SYNCPOINT test_named_syncpoints
                test_calls.append(name)  # pragma: SYNCPOINT test_named_syncpoints
                test_calls.append(name)  # pragma: SYNCPOINT

            return test

        test1 = make_test("A")
        test2 = make_test("B")

        permutations = orchestrate._permutations(["A", "B", "A", "B", "A", "B"])
        expected = list(itertools.chain(*permutations))

        counts = orchestrate.orchestrate(test1, test2, name="test_named_syncpoints")
        assert counts == (3, 3)
        assert test_calls == expected

    @staticmethod
    def test_syncpoints_decrease_after_initial_run():
        test_calls = []

        def make_test(name):
            syncpoints = [name] * 4

            def test():  # pragma: NO COVER
                test_calls.append(name)
                if syncpoints:
                    syncpoints.pop()  # pragma: SYNCPOINT
                    test_calls.append(name)

            return test

        test1 = make_test("A")
        test2 = make_test("B")

        expected = [
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "B",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
        ]

        counts = orchestrate.orchestrate(test1, test2)
        assert counts == (2, 2)
        assert test_calls == expected

    @staticmethod
    def test_syncpoints_increase_after_initial_run():
        test_calls = []

        def do_nothing():  # pragma: NO COVER
            pass

        def make_test(name):
            syncpoints = [None] * 4

            def test():  # pragma: NO COVER
                test_calls.append(name)  # pragma: SYNCPOINT
                test_calls.append(name)

                if syncpoints:
                    syncpoints.pop()
                else:
                    do_nothing()  # pragma: SYNCPOINT
                    test_calls.append(name)

            return test

        test1 = make_test("A")
        test2 = make_test("B")

        expected = [
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "B",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "A",
            "B",
            "B",
            "B",
            "A",
            "A",
            "A",
            "B",
        ]

        counts = orchestrate.orchestrate(test1, test2)
        assert counts == (2, 2)
        assert test_calls == expected

    @staticmethod
    def test_failure():
        test_calls = []

        def make_test(name):
            syncpoints = [None] * 4

            def test():  # pragma: NO COVER
                test_calls.append(name)  # pragma: SYNCPOINT
                test_calls.append(name)

                if syncpoints:
                    syncpoints.pop()
                else:
                    assert True is False

            return test

        test1 = make_test("A")
        test2 = make_test("B")

        expected = [
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "B",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
            "A",
            "B",
            "B",
            "A",
            "B",
            "A",
        ]

        with pytest.raises(AssertionError):
            orchestrate.orchestrate(test1, test2)

        assert test_calls == expected


def test__conductor():
    conductor = orchestrate._Conductor()
    items = []

    def run_in_test_thread():
        conductor.notify()
        items.append("test1")
        conductor.standby()
        items.append("test2")
        conductor.notify()
        conductor.standby()
        items.append("test3")
        conductor.notify()

    assert not items
    test_thread = threading.Thread(target=run_in_test_thread)

    test_thread.start()
    conductor.wait()
    assert items == ["test1"]

    conductor.go()
    conductor.wait()
    assert items == ["test1", "test2"]

    conductor.go()
    conductor.wait()
    assert items == ["test1", "test2", "test3"]


def test__get_syncpoints():  # pragma: SYNCPOINT test_get_syncpoints
    lines = enumerate(open(__file__, "r"), start=1)
    for expected_lineno, line in lines:  # pragma: NO BRANCH COVER
        if "# pragma: SYNCPOINT test_get_syncpoints" in line:
            break

    orchestrate._get_syncpoints(__file__)
    syncpoints = orchestrate._SYNCPOINTS[__file__]["test_get_syncpoints"]
    assert syncpoints == {expected_lineno}


class Test_TestThread:
    @staticmethod
    def test__sync():
        test_thread = orchestrate._TestThread(None, None)
        test_thread.conductor = mock.Mock()
        test_thread._sync()

        test_thread.conductor.notify.assert_called_once_with()
        test_thread.conductor.standby.assert_called_once_with()

    @staticmethod
    def test__trace_no_source_file():
        orchestrate._SYNCPOINTS.clear()
        frame = mock.Mock(f_globals={}, spec=("f_globals",))
        test_thread = orchestrate._TestThread(None, None)
        assert test_thread._trace(frame, None, None) is None
        assert not orchestrate._SYNCPOINTS

    @staticmethod
    def test__trace_this_source_file():
        orchestrate._SYNCPOINTS.clear()
        frame = mock.Mock(
            f_globals={"__file__": __file__},
            f_lineno=1,
            spec=(
                "f_globals",
                "f_lineno",
            ),
        )
        test_thread = orchestrate._TestThread(None, None)
        assert test_thread._trace(frame, None, None) == test_thread._trace
        assert __file__ in orchestrate._SYNCPOINTS

    @staticmethod
    def test__trace_reach_syncpoint():
        lines = enumerate(open(__file__, "r"), start=1)
        for syncpoint_lineno, line in lines:  # pragma: NO BRANCH COVER
            if "# pragma: SYNCPOINT test_get_syncpoints" in line:
                break

        orchestrate._SYNCPOINTS.clear()
        frame = mock.Mock(
            f_globals={"__file__": __file__},
            f_lineno=syncpoint_lineno,
            spec=(
                "f_globals",
                "f_lineno",
            ),
        )
        test_thread = orchestrate._TestThread(None, "test_get_syncpoints")
        test_thread._sync = mock.Mock()
        assert test_thread._trace(frame, None, None) == test_thread._trace
        test_thread._sync.assert_not_called()

        frame = mock.Mock(
            f_globals={"__file__": __file__},
            f_lineno=syncpoint_lineno + 1,
            spec=(
                "f_globals",
                "f_lineno",
            ),
        )
        assert test_thread._trace(frame, None, None) == test_thread._trace
        test_thread._sync.assert_called_once_with()

    @staticmethod
    def test__trace_other_source_file_with_no_syncpoints():
        filename = orchestrate.__file__
        if filename.endswith(".pyc"):  # pragma: NO COVER
            filename = filename[:-1]

        orchestrate._SYNCPOINTS.clear()
        frame = mock.Mock(
            f_globals={"__file__": filename + "c"},
            f_lineno=1,
            spec=(
                "f_globals",
                "f_lineno",
            ),
        )
        test_thread = orchestrate._TestThread(None, None)
        assert test_thread._trace(frame, None, None) is None
        syncpoints = orchestrate._SYNCPOINTS[filename]
        assert not syncpoints



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
from unittest import mock


def maybe_fail_import(predicate):
    """Create and return a patcher that conditionally makes an import fail.

    Args:
        predicate (Callable[[...], bool]): A callable that, if it returns `True`,
            triggers an `ImportError`. It must accept the same arguments as the
            built-in `__import__` function.
            https://docs.python.org/3/library/functions.html#__import__

    Returns:
        A mock patcher object that can be used to enable patched import behavior.
    """
    orig_import = builtins.__import__

    def custom_import(name, globals=None, locals=None, fromlist=(), level=0):
        if predicate(name, globals, locals, fromlist, level):
            raise ImportError
        return orig_import(name, globals, locals, fromlist, level)

    return mock.patch.object(builtins, "__import__", new=custom_import)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Tuple, Set

import click
from packaging.requirements import Requirement
from packaging.version import Version
import pkg_resources


def _get_package_requirements(package_name: str) -> List[Requirement]:
    """
    Get a list of all requirements and extras declared by this package.
    The package must already be installed in the environment.

    Args:
        package_name (str): The name of the package.

    Returns:
        List[pkg_resources.Requirement]: A list of package requirements and extras.
    """
    dist = pkg_resources.get_distribution(package_name)
    extras = tuple(dist.extras)
    requirements = [Requirement(str(r)) for r in dist.requires(extras=extras)]

    return requirements


def _parse_requirements_file(requirements_file: str) -> List[Requirement]:
    """
    Get a list of requirements found in a requirements file.

    Args:
        requirements_file (str): Path to a requirements file.

    Returns:
        List[Requirement]: A list of requirements.
    """
    requirements = []

    with Path(requirements_file).open() as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#"):
                requirements.append(Requirement(line))

    return requirements


def _get_pinned_versions(
    ctx: click.Context, requirements: List[Requirement]
) -> Set[Tuple[str, Version]]:
    """Turn a list of requirements into a set of (package name, Version) tuples.

    The requirements are all expected to pin explicitly to one version.
    Other formats will result in an error.

        {("requests", Version("1.25.0"), ("google-auth", Version("1.0.0")}

    Args:
        ctx (click.Context): The current click context.
        requirements (List[Requirement]): A list of requirements.

    Returns:
        Set[Tuple[str, Version]]: Tuples of the package name and Version.
    """
    constraints = set()

    invalid_requirements = []

    for constraint in requirements:
        spec_set = list(constraint.specifier)
        if len(spec_set) != 1:
            invalid_requirements.append(constraint.name)
        else:
            if spec_set[0].operator != "==":
                invalid_requirements.append(constraint.name)
            else:
                constraints.add((constraint.name, Version(spec_set[0].version)))

    if invalid_requirements:
        ctx.fail(
            f"These requirements are not pinned to one version: {invalid_requirements}"
        )

    return constraints


class IndeterminableLowerBound(Exception):
    pass


def _lower_bound(requirement: Requirement) -> str:
    """
    Given a requirement, determine the lowest version that fulfills the requirement.
    The lower bound can be determined for a requirement only if it is one of these
    formats:

        foo==1.2.0
        foo>=1.2.0
        foo>=1.2.0, <2.0.0dev
        foo<2.0.0dev, >=1.2.0

    Args:
        requirement (Requirement): A requirement to parse

    Returns:
        str: The lower bound for the requirement.
    """
    spec_set = list(requirement.specifier)

    # sort by operator: <, then >=
    spec_set.sort(key=lambda x: x.operator)

    if len(spec_set) == 1:
        # foo==1.2.0
        if spec_set[0].operator == "==":
            return spec_set[0].version
        # foo>=1.2.0
        elif spec_set[0].operator == ">=":
            return spec_set[0].version
    # foo<2.0.0, >=1.2.0 or foo>=1.2.0, <2.0.0
    elif len(spec_set) == 2:
        if spec_set[0].operator == "<" and spec_set[1].operator == ">=":
            return spec_set[1].version

    raise IndeterminableLowerBound(
        f"Lower bound could not be determined for {requirement.name}"
    )


def _get_package_lower_bounds(
    ctx: click.Context, requirements: List[Requirement]
) -> Set[Tuple[str, Version]]:
    """Get a set of tuples ('package_name', Version('1.0.0')) from a
    list of Requirements.

    Args:
        ctx (click.Context): The current click context.
        requirements (List[Requirement]): A list of requirements.

    Returns:
        Set[Tuple[str, Version]]: A set of (package_name, lower_bound)
            tuples.
    """
    bad_package_lower_bounds = []
    package_lower_bounds = set()

    for req in requirements:
        try:
            version = _lower_bound(req)
            package_lower_bounds.add((req.name, Version(version)))
        except IndeterminableLowerBound:
            bad_package_lower_bounds.append(req.name)

    if bad_package_lower_bounds:
        ctx.fail(
            f"setup.py is missing explicit lower bounds for the following packages: {str(bad_package_lower_bounds)}"
        )
    else:
        return package_lower_bounds


@click.group()
def main():
    pass


@main.command()
@click.option("--package-name", required=True, help="Name of the package.")
@click.option("--constraints-file", required=True, help="Path to constraints file.")
@click.pass_context
def update(ctx: click.Context, package_name: str, constraints_file: str) -> None:
    """Create a constraints file with lower bounds for package-name.

    If the constraints file already exists the contents will be overwritten.
    """
    requirements = _get_package_requirements(package_name)
    requirements.sort(key=lambda x: x.name)

    package_lower_bounds = list(_get_package_lower_bounds(ctx, requirements))
    package_lower_bounds.sort(key=lambda x: x[0])

    constraints = [f"{name}=={version}" for name, version in package_lower_bounds]
    Path(constraints_file).write_text("\n".join(constraints))


@main.command()
@click.option("--package-name", required=True, help="Name of the package.")
@click.option("--constraints-file", required=True, help="Path to constraints file.")
@click.pass_context
def check(ctx: click.Context, package_name: str, constraints_file: str):
    """Check that the constraints-file pins to the lower bound specified in package-name's
    setup.py for each requirement.

    Requirements:

    1. The setup.py pins every requirement in one of the following formats:

        * foo==1.2.0

        * foo>=1.2.0

        * foo>=1.2.0, <2.0.0dev

        * foo<2.0.0dev, >=1.2.0

    2. The constraints file pins every requirement to a single version:

        * foo==1.2.0

    3. package-name is already installed in the environment.
    """

    package_requirements = _get_package_requirements(package_name)
    constraints = _parse_requirements_file(constraints_file)

    package_lower_bounds = _get_package_lower_bounds(ctx, package_requirements)
    constraints_file_versions = _get_pinned_versions(ctx, constraints)

    # Look for dependencies in setup.py that are missing from constraints.txt
    package_names = {x[0] for x in package_lower_bounds}
    constraint_names = {x[0] for x in constraints_file_versions}
    missing_from_constraints = package_names - constraint_names

    if missing_from_constraints:
        ctx.fail(
            (
                f"The following packages are declared as a requirement or extra"
                f"in setup.py but were not found in {constraints_file}: {str(missing_from_constraints)}"
            )
        )

    # We use .issuperset() instead of == because there may be additional entries
    # in constraints.txt (e.g., test only requirements)
    if not constraints_file_versions.issuperset(package_lower_bounds):
        first_line = f"The following packages have different versions {package_name}'s setup.py and {constraints_file}"
        error_msg = [first_line, "-" * (7 + len(first_line))]

        difference = package_lower_bounds - constraints_file_versions
        constraints_dict = dict(constraints_file_versions)

        for req, setup_py_version in difference:
            error_msg.append(
                f"'{req}' lower bound is {setup_py_version} in setup.py but constraints file has {constraints_dict[req]}"
            )
        ctx.fail("\n".join(error_msg))

    click.secho("All good!", fg="green")


if __name__ == "__main__":
    main()



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import math
import queue
import sys
import threading
import tokenize


def orchestrate(*tests, **kwargs):
    """
    Orchestrate a deterministic concurrency test.

    Runs test functions in separate threads, with each thread taking turns running up
    until predefined syncpoints in a deterministic order. All possible orderings are
    tested.

    Most of the time, we try to use logic, best practices, and static analysis to insure
    correct operation of concurrent code. Sometimes our powers of reasoning fail us and,
    either through non-determistic stress testing or running code in production, a
    concurrent bug is discovered. When this occurs, we'd like to have a regression test
    to insure we've understood the problem and implemented a correct solution.
    `orchestrate` provides a means of deterministically testing concurrent code so we
    can write robust regression tests for complex concurrent scenarios.

    `orchestrate` runs each passed in test function in its own thread. Threads then
    "take turns" running. Turns are defined by setting syncpoints in the code under
    test, using comments containing "pragma: SYNCPOINT". `orchestrate` will scan the
    code under test and add syncpoints where it finds these comments.

    For example, let's say you have the following code in production::

        def hither_and_yon(destination):
            hither(destination)
            yon(destination)

    You've found there's a concurrency bug when two threads execute this code with the
    same argument, and you think that by adding a syncpoint between the calls to
    `hither` and `yon` you can reproduce the problem in a regression test. First add a
    comment with "pragma: SYNCPOINT" to the code under test::

        def hither_and_yon(destination):
            hither(destination)   # pragma: SYNCPOINT
            yon(destination)

    When testing with orchestrate, there will now be a syncpoint, or a pause, after the
    call to `hither` and before the call to `yon`. Now you can write a test to exercise
    `hither_and_yon` running in parallel::

        from unittest import mock
        from tests.unit import orchestrate

        from myorg.myproj.sales import travel

        def test_concurrent_hither_and_yon():

            def test_hither_and_yon():
                assert something
                travel.hither_and_yon("Raleigh")
                assert something_else

            counts = orchestrate.orchestrate(test_hither_and_yon, test_hither_and_yon)
            assert counts == (2, 2)

    What `orchestrate` will do now is take each of the two test functions passed in
    (actually the same function, twice, in this case), run them serially, and count the
    number of turns it takes to run each test to completion. In this example, it will
    take two turns for each test: one turn to start the thread and execute up until the
    syncpoint, and then another turn to execute from the syncpoint to the end of the
    test. The number of turns will always be one greater than the number of syncpoints
    encountered when executing the test.

    Once the counts have been taken, `orchestrate` will construct a test sequence that
    represents all of the turns taken by the passed in tests, with each value in the
    sequence representing the index of the test whose turn it is in the sequence. In
    this example, then, it would produce::

        [0, 0, 1, 1]

    This represents the first test taking both of its turns, followed by the second test
    taking both of its turns. At this point this scenario has already been tested,
    because this is what was run to produce the counts and the initial test sequence.
    Now `orchestrate` will run all of the remaining scenarios by finding all the
    permutations of the test sequence and executing those, in turn::

        [0, 1, 0, 1]
        [0, 1, 1, 0]
        [1, 0, 0, 1]
        [1, 0, 1, 0]
        [1, 1, 0, 0]

    You'll notice in our example that since both test functions are actually the same
    function, that although it tested 6 scenarios there are effectively only really 3
    unique scenarios. For the time being, though, `orchestrate` doesn't attempt to
    detect this condition or optimize for it.

    There are some performance considerations that should be taken into account when
    writing tests. The number of unique test sequences grows quite quickly with the
    number of turns taken by the functions under test. Our simple example with two
    threads each taking two turns, only yielded 6 scenarios, but two threads each taking
    6 turns, for example, yields 924 scenarios. Add another six step thread, for a total
    of three threads, and now you have over 17 thousand scenarios. In general, use the
    least number of steps/threads you can get away with and still expose the behavior
    you want to correct.

    For the same reason as above, it is recommended that if you have many concurrent
    tests, that you name your syncpoints so that you're not accidentally using
    syncpoints intended for other tests, as this will add steps to your tests. While
    it's not problematic from a testing standpoint to have extra steps in your tests, it
    can use computing resources unnecessarily. A name can be added to any syncpoint
    after the `SYNCPOINT` keyword in the pragma definition::

        def hither_and_yon(destination):
            hither(destination)   # pragma: SYNCPOINT hither and yon
            yon(destination)

    In your test, then, pass that name to `orchestrate` to cause it to use only
    syncpoints with that name::

        orchestrate.orchestrate(
            test_hither_and_yon, test_hither_and_yon, name="hither and yon"
        )

    As soon as any error or failure is detected, no more scenarios are run
    and that error is propagated to the main thread.

    One limitation of `orchestrate` is that it cannot really be used with `coverage`,
    since both tools use `sys.set_trace`. Any code that needs verifiable test coverage
    should have additional tests that do not use `orchestrate`, since code that is run
    under orchestrate will not show up in a coverage report generated by `coverage`.

    Args:
        tests (Tuple[Callable]): Test functions to be run. These functions will not be
            called with any arguments, so they must not have any required arguments.
        name (Optional[str]): Only use syncpoints with the given name. If omitted, only
            unnamed syncpoints will be used.

    Returns:
        Tuple[int]: A tuple of the count of the number turns for test passed in. Can be
            used a sanity check in tests to make sure you understand what's actually
            happening during a test.
    """
    name = kwargs.pop("name", None)
    if kwargs:
        raise TypeError(
            "Unexpected keyword arguments: {}".format(", ".join(kwargs.keys()))
        )

    # Produce an initial test sequence. The fundamental question we're always trying to
    # answer is "whose turn is it?" First we'll find out how many "turns" each test
    # needs to complete when run serially and use that to construct a sequence of
    # indexes. When a test's index appears in the sequence, it is that test's turn to
    # run. We'll start by constructing a sequence that would run each test through to
    # completion serially, one after the other.
    test_sequence = []
    counts = []
    for index, test in enumerate(tests):
        thread = _TestThread(test, name)
        for count in itertools.count(1):  # pragma: NO BRANCH
            # Pragma is required because loop never finishes naturally.
            thread.go()
            if thread.finished:
                break

        counts.append(count)
        test_sequence += [index] * count

    # Now we can take that initial sequence and generate all of its permutations,
    # running each one to try to uncover concurrency bugs
    sequences = iter(_permutations(test_sequence))

    # We already tested the first sequence getting our counts, so we can discard it
    next(sequences)

    # Test each sequence
    for test_sequence in sequences:
        threads = [_TestThread(test, name) for test in tests]
        try:
            for index in test_sequence:
                threads[index].go()

            # Its possible for number of turns to vary from one test run to the other,
            # especially if there is some undiscovered concurrency bug. Go ahead and
            # finish running each test to completion, if not already complete.
            for thread in threads:
                while not thread.finished:
                    thread.go()

        except Exception:
            # If an exception occurs, we still need to let any threads that are still
            # going finish up. Additional exceptions are silently ignored.
            for thread in threads:
                thread.finish()
            raise

    return tuple(counts)


_local = threading.local()


class _Conductor:
    """Coordinate communication between main thread and a test thread.

    Two way communicaton is maintained between the main thread and a test thread using
    two synchronized queues (`queue.Queue`) each with a size of one.
    """

    def __init__(self):
        self._notify = queue.Queue(1)
        self._go = queue.Queue(1)

    def notify(self):
        """Called from test thread to let us know it's finished or is ready for its next
        turn."""
        self._notify.put(None)

    def standby(self):
        """Called from test thread in order to block until told to go."""
        self._go.get()

    def wait(self):
        """Called from main thread to wait for test thread to either get to the
        next syncpoint or finish."""
        self._notify.get()

    def go(self):
        """Called from main thread to tell test thread to go."""
        self._go.put(None)


_SYNCPOINTS = {}
"""Dict[str, Dict[str, Set[int]]]: Dict mapping source fileneme to a dict mapping
syncpoint name to set of line numbers where syncpoints with that name occur in the
source file.
"""


def _get_syncpoints(filename):
    """Find syncpoints in a source file.

    Does a simple tokenization of the source file, looking for comments with "pragma:
    SYNCPOINT", and populates _SYNCPOINTS using the syncpoint name and line number in
    the source file.
    """
    _SYNCPOINTS[filename] = syncpoints = {}

    # Use tokenize to find pragma comments
    with open(filename, "r") as pyfile:
        tokens = tokenize.generate_tokens(pyfile.readline)
        for type, value, start, end, line in tokens:
            if type == tokenize.COMMENT and "pragma: SYNCPOINT" in value:
                name = value.split("SYNCPOINT", 1)[1].strip()
                if not name:
                    name = None

                if name not in syncpoints:
                    syncpoints[name] = set()

                lineno, column = start
                syncpoints[name].add(lineno)


class _TestThread:
    """A thread for a test function."""

    thread = None
    finished = False
    error = None
    at_syncpoint = False

    def __init__(self, test, name):
        self.test = test
        self.name = name
        self.conductor = _Conductor()

    def _run(self):
        sys.settrace(self._trace)
        _local.conductor = self.conductor
        try:
            self.test()
        except Exception as error:
            self.error = error
        finally:
            self.finished = True
            self.conductor.notify()

    def _sync(self):
        # Tell main thread we're finished, for now
        self.conductor.notify()

        # Wait for the main thread to tell us to go again
        self.conductor.standby()

    def _trace(self, frame, event, arg):
        """Argument to `sys.settrace`.

        Handles frames during test run, syncing at syncpoints, when found.

        Returns:
            `None` if no more tracing is required for the function call, `self._trace`
            if tracing should continue.
        """
        if self.at_syncpoint:
            # We hit a syncpoint on the previous call, so now we sync.
            self._sync()
            self.at_syncpoint = False

        filename = frame.f_globals.get("__file__")
        if not filename:
            # Can't trace code without a source file
            return

        if filename.endswith(".pyc"):
            filename = filename[:-1]

        if filename not in _SYNCPOINTS:
            _get_syncpoints(filename)

        syncpoints = _SYNCPOINTS[filename].get(self.name)
        if not syncpoints:
            # This file doesn't contain syncpoints, don't continue to trace
            return

        # We've hit a syncpoint. Execute whatever line the syncpoint is on and then
        # sync next time this gets called.
        if frame.f_lineno in syncpoints:
            self.at_syncpoint = True

        return self._trace

    def go(self):
        if self.finished:
            return

        if self.thread is None:
            self.thread = threading.Thread(target=self._run)
            self.thread.start()

        else:
            self.conductor.go()

        self.conductor.wait()

        if self.error:
            raise self.error

    def finish(self):
        while not self.finished:
            try:
                self.go()
            except Exception:
                pass


class _permutations:
    """Generates a sequence of all permutations of `sequence`.

    Permutations are returned in lexicographic order using the "Generation in
    lexicographic order" algorithm described in `the Wikipedia article on "Permutation"
    <https://en.wikipedia.org/wiki/Permutation>`_.

    This implementation differs significantly from `itertools.permutations` in that the
    value of individual elements is taken into account, thus eliminating redundant
    orderings that would be produced by `itertools.permutations`.

    Args:
        sequence (Sequence[Any]): Sequence must be finite and orderable.

    Returns:
        Sequence[Sequence[Any]]: Set of all permutations of `sequence`.
    """

    def __init__(self, sequence):
        self._start = tuple(sorted(sequence))

    def __len__(self):
        """Compute the number of permutations.

        Let the number of elements in a sequence N and the number of repetitions for
        individual members of the sequence be n1, n2, ... nx. The number of unique
        permutations is: N! / n1! / n2! / ... / nx!.

        For example, let `sequence` be [1, 2, 3, 1, 2, 3, 1, 2, 3]. The number of unique
        permutations is: 9! / 3! / 3! / 3! = 1680.

        See: "Permutations of multisets" in `the Wikipedia article on "Permutation"
        <https://en.wikipedia.org/wiki/Permutation>`_.
        """
        repeats = [len(list(group)) for value, group in itertools.groupby(self._start)]
        length = math.factorial(len(self._start))
        for repeat in repeats:
            length /= math.factorial(repeat)

        return int(length)

    def __iter__(self):
        """Iterate over permutations.

        See: "Generation in lexicographic order" algorithm described in `the Wikipedia
        article on "Permutation" <https://en.wikipedia.org/wiki/Permutation>`_.
        """
        current = list(self._start)
        size = len(current)

        while True:
            yield tuple(current)

            # 1. Find the largest index i such that a[i] < a[i + 1].
            for i in range(size - 2, -1, -1):
                if current[i] < current[i + 1]:
                    break

            else:
                # If no such index exists, the permutation is the last permutation.
                return

            # 2. Find the largest index j greater than i such that a[i] < a[j].
            for j in range(size - 1, i, -1):
                if current[i] < current[j]:
                    break

            else:  # pragma: NO COVER
                raise RuntimeError("Broken algorithm")

            # 3. Swap the value of a[i] with that of a[j].
            temp = current[i]
            current[i] = current[j]
            current[j] = temp

            # 4. Reverse the sequence from a[i + 1] up to and including the final
            # element a[n].
            current = current[: i + 1] + list(reversed(current[i + 1 :]))



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import time

MAX_TRIES = 4
DELAY = 1
BACKOFF = 2


def _retry_all(_):
    """Retry all caught exceptions."""
    return True


class BackoffFailed(Exception):
    """Retry w/ backoffs did not complete successfully."""


class RetryBase(object):
    """Base for retrying calling a decorated function w/ exponential backoff.

    :type max_tries: int
    :param max_tries: Number of times to try (not retry) before giving up.

    :type delay: int
    :param delay: Initial delay between retries in seconds.

    :type backoff: int
    :param backoff: Backoff multiplier e.g. value of 2 will double the
                    delay each retry.

    :type logger: logging.Logger instance
    :param logger: Logger to use. If None, print.
    """

    def __init__(self, max_tries=MAX_TRIES, delay=DELAY, backoff=BACKOFF, logger=None):
        self.max_tries = max_tries
        self.delay = delay
        self.backoff = backoff
        self.logger = logger.warning if logger else print


class RetryErrors(RetryBase):
    """Decorator for retrying given exceptions in testing.

    :type exception: Exception or tuple of Exceptions
    :param exception: The exception to check or may be a tuple of
                      exceptions to check.

    :type error_predicate: function, takes caught exception, returns bool
    :param error_predicate: Predicate evaluating whether to retry after a
                            caught exception.

    :type max_tries: int
    :param max_tries: Number of times to try (not retry) before giving up.

    :type delay: int
    :param delay: Initial delay between retries in seconds.

    :type backoff: int
    :param backoff: Backoff multiplier e.g. value of 2 will double the
                    delay each retry.

    :type logger: logging.Logger instance
    :param logger: Logger to use. If None, print.
    """

    def __init__(
        self,
        exception,
        error_predicate=_retry_all,
        max_tries=MAX_TRIES,
        delay=DELAY,
        backoff=BACKOFF,
        logger=None,
    ):
        super(RetryErrors, self).__init__(max_tries, delay, backoff, logger)
        self.exception = exception
        self.error_predicate = error_predicate

    def __call__(self, to_wrap):
        @functools.wraps(to_wrap)
        def wrapped_function(*args, **kwargs):
            tries = 0
            while tries < self.max_tries:
                try:
                    return to_wrap(*args, **kwargs)
                except self.exception as caught_exception:

                    if not self.error_predicate(caught_exception):
                        raise

                    delay = self.delay * self.backoff ** tries
                    msg = "%s, Trying again in %d seconds..." % (
                        caught_exception,
                        delay,
                    )
                    self.logger(msg)

                    time.sleep(delay)
                    tries += 1
            return to_wrap(*args, **kwargs)

        return wrapped_function


class RetryResult(RetryBase):
    """Decorator for retrying based on non-error result.

    :type result_predicate: function, takes result, returns bool
    :param result_predicate: Predicate evaluating whether to retry after a
                             result is returned.

    :type max_tries: int
    :param max_tries: Number of times to try (not retry) before giving up.

    :type delay: int
    :param delay: Initial delay between retries in seconds.

    :type backoff: int
    :param backoff: Backoff multiplier e.g. value of 2 will double the
                    delay each retry.

    :type logger: logging.Logger instance
    :param logger: Logger to use. If None, print.
    """

    def __init__(
        self,
        result_predicate,
        max_tries=MAX_TRIES,
        delay=DELAY,
        backoff=BACKOFF,
        logger=None,
    ):
        super(RetryResult, self).__init__(max_tries, delay, backoff, logger)
        self.result_predicate = result_predicate

    def __call__(self, to_wrap):
        @functools.wraps(to_wrap)
        def wrapped_function(*args, **kwargs):
            tries = 0
            while tries < self.max_tries:
                result = to_wrap(*args, **kwargs)
                if self.result_predicate(result):
                    return result

                delay = self.delay * self.backoff ** tries
                msg = "%s. Trying again in %d seconds..." % (
                    self.result_predicate.__name__,
                    delay,
                )
                self.logger(msg)

                time.sleep(delay)
                tries += 1
            raise BackoffFailed()

        return wrapped_function


class RetryInstanceState(RetryBase):
    """Decorator for retrying based on instance state.

    :type instance_predicate: function, takes instance, returns bool
    :param instance_predicate: Predicate evaluating whether to retry after an
                               API-invoking method is called.

    :type max_tries: int
    :param max_tries: Number of times to try (not retry) before giving up.

    :type delay: int
    :param delay: Initial delay between retries in seconds.

    :type backoff: int
    :param backoff: Backoff multiplier e.g. value of 2 will double the
                    delay each retry.

    :type logger: logging.Logger instance
    :param logger: Logger to use. If None, print.
    """

    def __init__(
        self,
        instance_predicate,
        max_tries=MAX_TRIES,
        delay=DELAY,
        backoff=BACKOFF,
        logger=None,
    ):
        super(RetryInstanceState, self).__init__(max_tries, delay, backoff, logger)
        self.instance_predicate = instance_predicate

    def __call__(self, to_wrap):
        instance = to_wrap.__self__  # only instance methods allowed

        @functools.wraps(to_wrap)
        def wrapped_function(*args, **kwargs):
            tries = 0
            while tries < self.max_tries:
                result = to_wrap(*args, **kwargs)
                if self.instance_predicate(instance):
                    return result

                delay = self.delay * self.backoff ** tries
                msg = "%s. Trying again in %d seconds..." % (
                    self.instance_predicate.__name__,
                    delay,
                )
                self.logger(msg)

                time.sleep(delay)
                tries += 1
            raise BackoffFailed()

        return wrapped_function



# Copyright 2014 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import os
import sys
import time

import google.auth.credentials  # type: ignore
from google.auth.environment_vars import CREDENTIALS as TEST_CREDENTIALS  # type: ignore


# From shell environ. May be None.
CREDENTIALS = os.getenv(TEST_CREDENTIALS)

ENVIRON_ERROR_MSG = """\
To run the system tests, you need to set some environment variables.
Please check the CONTRIBUTING guide for instructions.
"""


class EmulatorCreds(google.auth.credentials.Credentials):
    """A mock credential object.

    Used to avoid unnecessary token refreshing or reliance on the network
    while an emulator is running.
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        self.token = b"seekrit"
        self.expiry = None

    @property
    def valid(self):
        """Would-be validity check of the credentials.

        Always is :data:`True`.
        """
        return True

    def refresh(self, unused_request):  # pylint: disable=unused-argument
        """Off-limits implementation for abstract method."""
        raise RuntimeError("Should never be refreshed.")


def check_environ():
    err_msg = None
    if CREDENTIALS is None:
        err_msg = "\nMissing variables: " + TEST_CREDENTIALS
    elif not os.path.isfile(CREDENTIALS):
        err_msg = "\nThe %s path %r is not a file." % (TEST_CREDENTIALS, CREDENTIALS)

    if err_msg is not None:
        msg = ENVIRON_ERROR_MSG + err_msg
        print(msg, file=sys.stderr)
        sys.exit(1)


def unique_resource_id(delimiter="_"):
    """A unique identifier for a resource.

    Intended to help locate resources created in particular
    testing environments and at particular times.
    """
    build_id = os.getenv("CIRCLE_BUILD_NUM", "")
    if build_id == "":
        return "%s%d" % (delimiter, 1000 * time.time())
    else:
        return "%s%s%s%d" % (delimiter, build_id, delimiter, time.time())



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import random
import re

from typing import Union

_RESOURCE_DATE_FORMAT = "%Y%m%d%H%M%S"
_RESOURCE_DATE_LENGTH = 4 + 2 + 2 + 2 + 2 + 2
_RE_SEPARATORS = re.compile(r"[/\-\\_]")


def _common_prefix(repo, relative_dir, separator="_"):
    repo = _RE_SEPARATORS.sub(separator, repo)
    relative_dir = _RE_SEPARATORS.sub(separator, relative_dir)
    return f"{repo}{separator}{relative_dir}"


class Prefixer(object):
    """Create/manage resource IDs for system testing.

    Usage:

    Creating resources:

    >>> import test_utils.prefixer
    >>> prefixer = test_utils.prefixer.Prefixer("python-bigquery", "samples/snippets")
    >>> dataset_id = prefixer.create_prefix() + "my_sample"

    Cleaning up resources:

    >>> @pytest.fixture(scope="session", autouse=True)
    ... def cleanup_datasets(bigquery_client: bigquery.Client):
    ...     for dataset in bigquery_client.list_datasets():
    ...         if prefixer.should_cleanup(dataset.dataset_id):
    ...             bigquery_client.delete_dataset(
    ...                 dataset, delete_contents=True, not_found_ok=True
    """

    def __init__(
        self, repo, relative_dir, separator="_", cleanup_age=datetime.timedelta(days=1)
    ):
        self._separator = separator
        self._cleanup_age = cleanup_age
        self._prefix = _common_prefix(repo, relative_dir, separator=separator)

    def create_prefix(self) -> str:
        timestamp = datetime.datetime.utcnow().strftime(_RESOURCE_DATE_FORMAT)
        random_string = hex(random.randrange(0x1000000))[2:]
        return f"{self._prefix}{self._separator}{timestamp}{self._separator}{random_string}"

    def _name_to_date(self, resource_name: str) -> Union[datetime.datetime, None]:
        start_date = len(self._prefix) + len(self._separator)
        date_string = resource_name[start_date : start_date + _RESOURCE_DATE_LENGTH]
        try:
            parsed_date = datetime.datetime.strptime(date_string, _RESOURCE_DATE_FORMAT)
            return parsed_date
        except ValueError:
            return None

    def should_cleanup(self, resource_name: str) -> bool:
        yesterday = datetime.datetime.utcnow() - self._cleanup_age
        if not resource_name.startswith(self._prefix):
            return False

        created_date = self._name_to_date(resource_name)
        return created_date is not None and created_date < yesterday



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools


# This package has four requirements.
# Each uses a different kind of pin accepted by the function that
# extracts lower bounds.
requirements = [
    "requests>=1.0.0",
    "packaging>=14.0, <22.0.0",
    "wheel<0.36.0, >=0.35.0",
    "click==7.0.0",
]

extras = {"grpc": "grpcio>=1.0.0"}

setuptools.setup(
    name="valid-package",
    version="0.0.1",
    author="Example Author",
    author_email="author@example.com",
    description="A small example package",
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    extras_require=extras,
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from pathlib import Path
import re
import tempfile
from typing import List

from click.testing import CliRunner
import pytest  # type: ignore

from test_utils.lower_bound_checker import lower_bound_checker

RUNNER = CliRunner()

PACKAGE_LIST_REGEX = re.compile(r"Error.*[\[\{](.+)[\]\}]")
DIFFERENT_VERSIONS_LIST_REGEX = re.compile("'(.*?)' lower bound is")

# These packages are installed into the environment by the nox session
# See 'resources/' for the setup.py files
GOOD_PACKAGE = "valid-package"
BAD_PACKAGE = "invalid-package"


def parse_error_msg(msg: str) -> List[str]:
    """Get package names from the error message.
    
    Example:
        Error: setup.py is missing explicit lower bounds for the following packages: ["requests", "grpcio"]
    """
    match = PACKAGE_LIST_REGEX.search(msg)

    reqs: List[str] = []

    if match:
        reqs = match.groups(1)[0].split(",")  # type: ignore
        reqs = [r.strip().replace("'", "").replace('"', "") for r in reqs]
    
    return reqs

def parse_diff_versions_error_msg(msg: str) -> List[str]:
    """Get package names from the error message listing different versions
    
    Example:
        'requests' lower bound is 1.2.0 in setup.py but constraints file has 1.3.0
        'grpcio' lower bound is 1.0.0 in setup.py but constraints file has 1.10.0
    """
    pattern = re.compile(DIFFERENT_VERSIONS_LIST_REGEX)
    pkg_names = pattern.findall(msg)

    return pkg_names

@contextmanager
def constraints_file(requirements: List[str]):
    """Write the list of requirements into a temporary file"""

    tmpdir = tempfile.TemporaryDirectory()
    constraints_path = Path(tmpdir.name) / "constraints.txt"

    constraints_path.write_text("\n".join(requirements))
    yield constraints_path

    tmpdir.cleanup()


def test_update_constraints():
    with tempfile.TemporaryDirectory() as tmpdir:
        constraints_path = Path(tmpdir) / "constraints.txt"

        result = RUNNER.invoke(
            lower_bound_checker.update, ["--package-name", GOOD_PACKAGE, "--constraints-file", str(constraints_path)]
        )

        assert result.exit_code == 0
        assert constraints_path.exists()
    
        output = constraints_path.read_text().split("\n")

        assert output == ["click==7.0.0", "grpcio==1.0.0", "packaging==14.0", "requests==1.0.0", "wheel==0.35.0",]



def test_update_constraints_overwrites_existing_file():
    constraints = [
        "requests==1.0.0",
        "packaging==13.0",
        "wheel==0.36.0",
        "click==5.0.0",
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.update, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

        assert result.exit_code == 0

        output = c.read_text().split("\n")
        assert output == ["click==7.0.0", "grpcio==1.0.0", "packaging==14.0", "requests==1.0.0", "wheel==0.35.0",
        ]

def test_update_constraints_with_setup_py_missing_lower_bounds():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        "wheel==0.35.0",
        "click==7.0.0",
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.update, ["--package-name", BAD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 2
    assert "setup.py is missing explicit lower bounds" in result.output

    invalid_pkg_list = parse_error_msg(result.output)
    assert set(invalid_pkg_list) == {"requests", "packaging", "wheel"}



def test_check():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        "wheel==0.35.0",
        "click==7.0.0",
        "grpcio==1.0.0"
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 0


def test_update_constraints_with_extra_constraints():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        "wheel==0.35.0",
        "click==7.0.0",
        "grpcio==1.0.0",
        "pytest==6.0.0",  # additional requirement
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 0


def test_check_with_missing_constraints_file():
    result = RUNNER.invoke(
        lower_bound_checker.check,
        [
            "--package-name",
            GOOD_PACKAGE,
            "--constraints-file",
            "missing_constraints.txt",
        ],
    )

    assert result.exit_code == 1
    assert isinstance(result.exception, FileNotFoundError)


def test_check_with_constraints_file_invalid_pins():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        "wheel==1.0.0, <2.0.0dev",  # should be ==
        "click>=7.0.0",  # should be ==
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 2

    invalid_pkg_list = parse_error_msg(result.output)

    assert set(invalid_pkg_list) == {"wheel", "click"}


def test_check_with_constraints_file_missing_packages():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        # missing 'wheel' and 'click' and extra 'grpcio'
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 2

    invalid_pkg_list = parse_error_msg(result.output)
    assert set(invalid_pkg_list) == {"wheel", "click", "grpcio"}


def test_check_with_constraints_file_different_versions():
    constraints = [
        "requests==1.2.0",  # setup.py has 1.0.0
        "packaging==14.1",  # setup.py has 14.0
        "wheel==0.36.0",  # setup.py has 0.35.0
        "click==7.0.0",
        "grpcio==1.0.0"
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", GOOD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 2

    invalid_pkg_list = parse_diff_versions_error_msg(result.output)
    assert set(invalid_pkg_list) == {"requests", "packaging", "wheel"}


def test_check_with_setup_py_missing_lower_bounds():
    constraints = [
        "requests==1.0.0",
        "packaging==14.0",
        "wheel==1.0.0",
        "click==7.0.0",
    ]
    with constraints_file(constraints) as c:
        result = RUNNER.invoke(
            lower_bound_checker.check, ["--package-name", BAD_PACKAGE, "--constraints-file", c]
        )

    assert result.exit_code == 2

    invalid_pkg_list = parse_error_msg(result.output)
    assert set(invalid_pkg_list) == {"requests", "packaging", "wheel"}



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re

import pytest  # type: ignore

import test_utils.prefixer


class FakeDateTime(object):
    """Fake datetime class since pytest can't monkeypatch attributes of
    built-in/extension type.
    """

    def __init__(self, fake_now):
        self._fake_now = fake_now

    def utcnow(self):
        return self._fake_now

    strptime = datetime.datetime.strptime


@pytest.mark.parametrize(
    ("repo", "relative_dir", "separator", "expected"),
    [
        (
            "python-bigquery",
            "samples/snippets",
            "_",
            "python_bigquery_samples_snippets",
        ),
        ("python-storage", "samples\\snippets", "-", "python-storage-samples-snippets"),
    ],
)
def test_common_prefix(repo, relative_dir, separator, expected):
    got = test_utils.prefixer._common_prefix(repo, relative_dir, separator=separator)
    assert got == expected


def test_create_prefix(monkeypatch):
    fake_datetime = FakeDateTime(datetime.datetime(2021, 6, 21, 3, 32, 0))
    monkeypatch.setattr(datetime, "datetime", fake_datetime)

    prefixer = test_utils.prefixer.Prefixer(
        "python-test-utils", "tests/unit", separator="?"
    )
    got = prefixer.create_prefix()
    parts = got.split("?")
    assert len(parts) == 7
    assert "?".join(parts[:5]) == "python?test?utils?tests?unit"
    datetime_part = parts[5]
    assert datetime_part == "20210621033200"
    random_hex_part = parts[6]
    assert re.fullmatch("[0-9a-f]+", random_hex_part)


@pytest.mark.parametrize(
    ("resource_name", "separator", "expected"),
    [
        ("test_utils_created_elsewhere", "_", False),
        ("test_utils_20210620120000", "_", False),
        ("test_utils_20210620120000_abcdef_my_name", "_", False),
        ("test_utils_20210619120000", "_", True),
        ("test_utils_20210619120000_abcdef_my_name", "_", True),
        ("test?utils?created?elsewhere", "_", False),
        ("test?utils?20210620120000", "?", False),
        ("test?utils?20210619120000", "?", True),
    ],
)
def test_should_cleanup(resource_name, separator, expected, monkeypatch):
    fake_datetime = FakeDateTime(datetime.datetime(2021, 6, 21, 3, 32, 0))
    monkeypatch.setattr(datetime, "datetime", fake_datetime)

    prefixer = test_utils.prefixer.Prefixer("test", "utils", separator=separator)
    assert prefixer.should_cleanup(resource_name) == expected



# -*- coding: utf-8 -*-
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest  # type: ignore


INSIDE_VPCSC_ENVVAR = "GOOGLE_CLOUD_TESTS_IN_VPCSC"
PROJECT_INSIDE_ENVVAR = "PROJECT_ID"
PROJECT_OUTSIDE_ENVVAR = "GOOGLE_CLOUD_TESTS_VPCSC_OUTSIDE_PERIMETER_PROJECT"
BUCKET_OUTSIDE_ENVVAR = "GOOGLE_CLOUD_TESTS_VPCSC_OUTSIDE_PERIMETER_BUCKET"


class VPCSCTestConfig(object):
    """System test utility for VPCSC detection.

    See: https://cloud.google.com/vpc-service-controls/docs/
    """

    @property
    def inside_vpcsc(self):
        """Test whether the test environment is configured to run inside VPCSC.

        Returns:
            bool:
                true if the environment is configured to run inside VPCSC,
                else false.
        """
        return INSIDE_VPCSC_ENVVAR in os.environ

    @property
    def project_inside(self):
        """Project ID for testing outside access.

        Returns:
            str: project ID used for testing outside access; None if undefined.
        """
        return os.environ.get(PROJECT_INSIDE_ENVVAR, None)

    @property
    def project_outside(self):
        """Project ID for testing inside access.

        Returns:
            str: project ID used for testing inside access; None if undefined.
        """
        return os.environ.get(PROJECT_OUTSIDE_ENVVAR, None)

    @property
    def bucket_outside(self):
        """GCS bucket for testing inside access.

        Returns:
            str: bucket ID used for testing inside access; None if undefined.
        """
        return os.environ.get(BUCKET_OUTSIDE_ENVVAR, None)

    def skip_if_inside_vpcsc(self, testcase):
        """Test decorator: skip if running inside VPCSC."""
        reason = (
            "Running inside VPCSC.  "
            "Unset the {} environment variable to enable this test."
        ).format(INSIDE_VPCSC_ENVVAR)
        skip = pytest.mark.skipif(self.inside_vpcsc, reason=reason)
        return skip(testcase)

    def skip_unless_inside_vpcsc(self, testcase):
        """Test decorator: skip if running outside VPCSC."""
        reason = (
            "Running outside VPCSC.  "
            "Set the {} environment variable to enable this test."
        ).format(INSIDE_VPCSC_ENVVAR)
        skip = pytest.mark.skipif(not self.inside_vpcsc, reason=reason)
        return skip(testcase)

    def skip_unless_inside_project(self, testcase):
        """Test decorator: skip if inside project env var not set."""
        reason = (
            "Project ID for running inside VPCSC not set.  "
            "Set the {} environment variable to enable this test."
        ).format(PROJECT_INSIDE_ENVVAR)
        skip = pytest.mark.skipif(self.project_inside is None, reason=reason)
        return skip(testcase)

    def skip_unless_outside_project(self, testcase):
        """Test decorator: skip if outside project env var not set."""
        reason = (
            "Project ID for running outside VPCSC not set. "
            "Set the {} environment variable to enable this test."
        ).format(PROJECT_OUTSIDE_ENVVAR)
        skip = pytest.mark.skipif(self.project_outside is None, reason=reason)
        return skip(testcase)

    def skip_unless_outside_bucket(self, testcase):
        """Test decorator: skip if outside bucket env var not set."""
        reason = (
            "Bucket ID for running outside VPCSC not set. "
            "Set the {} environment variable to enable this test."
        ).format(BUCKET_OUTSIDE_ENVVAR)
        skip = pytest.mark.skipif(self.bucket_outside is None, reason=reason)
        return skip(testcase)


vpcsc_config = VPCSCTestConfig()
